"use strict";
/**
 * The Iterable module provides tools for working with Typescript's Iterable<T> type in a functional way.
 *
 * In functional jargon, this module provides a monadic interface over Typescript's Iterable<T>.
 *
 * This nodule does not implements fuctions that would lead to degrading perfomance of Iterable. Any Iterable can be easly converted to Array
 *
 * @since 1.0.0
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || (function () {
    var ownKeys = function(o) {
        ownKeys = Object.getOwnPropertyNames || function (o) {
            var ar = [];
            for (var k in o) if (Object.prototype.hasOwnProperty.call(o, k)) ar[ar.length] = k;
            return ar;
        };
        return ownKeys(o);
    };
    return function (mod) {
        if (mod && mod.__esModule) return mod;
        var result = {};
        if (mod != null) for (var k = ownKeys(mod), i = 0; i < k.length; i++) if (k[i] !== "default") __createBinding(result, mod, k[i]);
        __setModuleDefault(result, mod);
        return result;
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
exports.toArray = exports.foldMap = exports.foldMapWithIndex = exports.reduce = exports.reduceWithIndex = exports.apS = exports.let = exports.bindTo = exports.bind = exports.Do = exports.tapIO = exports.tap = exports.uniq = exports.lefts = exports.rights = exports.compact = exports.filterMap = exports.filterMapWithIndex = exports.filter = exports.filterWithIndex = exports.MonadIO = exports.FromIO = exports.Monad = exports.Unfoldable = exports.Chain = exports.flatten = exports.flatMap = exports.Applicative = exports.apSecond = exports.apFirst = exports.Apply = exports.ap = exports.flap = exports.asUnit = exports.as = exports.FunctorWithIndex = exports.Functor = exports.map = exports.mapWithIndex = exports.Pointed = exports.of = exports.fromReadonlyArray = exports.fromIO = exports.unfold = exports.makeBy = exports.makeByWithIndex = exports.URI = void 0;
exports.transform = transform;
const Apply_1 = require("fp-ts/lib/Apply");
const chainable = __importStar(require("fp-ts/lib/Chain"));
const Functor_1 = require("fp-ts/lib/Functor");
const O = __importStar(require("fp-ts/lib/Option"));
const function_1 = require("fp-ts/lib/function");
const internal_1 = require("./internal");
/**
 * @category type lambdas
 * @since 1.0.0
 */
exports.URI = "Iterable";
/**
 * Return a `Iterable` with elements initialized with `f(i)`.
 * Iterable stops when f return O.none
 *
 * @category constructors
 * @since 1.0.0
 */
const makeByWithIndex = (f) => ({
    *[Symbol.iterator]() {
        for (let i = 0, el = f(i); O.isSome(el); i++, el = f(i)) {
            yield el.value;
        }
    },
});
exports.makeByWithIndex = makeByWithIndex;
/**
 * @category constructors
 * @since 1.0.0
 */
const makeBy = (f) => (0, exports.makeByWithIndex)(() => f());
exports.makeBy = makeBy;
/**
 * @category constructors
 * @since 1.0.0
 */
const unfold = (b, f) => {
    let bb = b;
    return {
        *[Symbol.iterator]() {
            for (let mt = f(bb); O.isSome(mt); mt = f(bb)) {
                const [a, b] = mt.value;
                bb = b;
                yield a;
            }
        },
    };
};
exports.unfold = unfold;
/**
 * @category conversions
 * @since 1.0.0
 */
const fromIO = (ma) => ({
    *[Symbol.iterator]() {
        yield ma();
    },
});
exports.fromIO = fromIO;
/**
 * @category conversions
 * @since 1.0.0
 * isn't it weird? all arrays are iterables
 */
const fromReadonlyArray = (ax) => ({
    *[Symbol.iterator]() {
        for (const a of ax) {
            yield a;
        }
    },
});
exports.fromReadonlyArray = fromReadonlyArray;
/**
 * @category constructors
 * @since 1.0.0
 */
const of = (a) => ({
    *[Symbol.iterator]() {
        yield a;
    },
});
exports.of = of;
/**
 * @category instances
 * @since 1.0.0
 */
exports.Pointed = { URI: exports.URI, of: exports.of };
/**
 * @category mapping
 * @since 1.0.0
 */
const mapWithIndex = (f) => (fa) => {
    let i = 0;
    return {
        *[Symbol.iterator]() {
            for (const a of fa) {
                yield f(i++, a);
            }
        },
    };
};
exports.mapWithIndex = mapWithIndex;
/**
 * @category mapping
 * @since 1.0.0
 */
const map = (f) => (0, exports.mapWithIndex)((_i, a) => f(a));
exports.map = map;
/**
 * @category instances
 * @since 1.0.0
 */
exports.Functor = {
    URI: exports.URI,
    map: (fa, f) => (0, exports.map)(f)(fa),
};
/**
 * @category instances
 * @since 1.0.0
 */
exports.FunctorWithIndex = {
    ...exports.Functor,
    mapWithIndex: (fa, f) => (0, exports.mapWithIndex)(f)(fa),
};
/**
 * Maps the value to the specified constant value.
 *
 * @category mapping
 * @since 1.0.0
 */
exports.as = (0, internal_1.dual)(2, (0, internal_1.as)(exports.Functor));
/**
 * Maps the value to the void constant value.
 *
 * @category mapping
 * @since 1.0.0
 */
exports.asUnit = (0, internal_1.asUnit)(exports.Functor);
/**
 * @category mapping
 * @since 1.0.0
 */
exports.flap = (0, Functor_1.flap)(exports.Functor);
/**
 * @category apply
 * @since 1.0.0
 */
const ap = (fa) => (fab) => ({
    *[Symbol.iterator]() {
        for (const ab of fab) {
            for (const a of fa) {
                yield ab(a);
            }
        }
    },
});
exports.ap = ap;
/**
 * @category instances
 * @since 1.0.0
 */
exports.Apply = {
    ...exports.Functor,
    ap: (fab, fa) => (0, exports.ap)(fa)(fab),
};
/**
 * @category apply
 * @since 1.0.0
 */
exports.apFirst = (0, Apply_1.apFirst)(exports.Apply);
/**
 * @category apply
 * @since 1.0.0
 */
exports.apSecond = (0, Apply_1.apSecond)(exports.Apply);
/**
 * @category instances
 * @since 1.0.0
 */
exports.Applicative = {
    ...exports.Pointed,
    ...exports.Apply,
};
/**
 * @category sequencing
 * @since 1.0.0
 */
const flatMap = (f) => (fa) => ({
    *[Symbol.iterator]() {
        for (const a1 of fa) {
            for (const a2 of f(a1)) {
                yield a2;
            }
        }
    },
});
exports.flatMap = flatMap;
/**
 * @category sequencing
 * @since 1.0.0
 */
exports.flatten = 
/*#__PURE__*/ (0, exports.flatMap)(function_1.identity);
/**
 * @category instances
 * @since 1.0.0
 */
exports.Chain = {
    ...exports.Apply,
    chain: (fab, fa) => (0, exports.flatMap)(fa)(fab),
};
/**
 * @category instances
 * @since 1.0.0
 */
exports.Unfoldable = {
    URI: exports.URI,
    unfold: exports.unfold,
};
/**
 * @category instances
 * @since 1.0.0
 */
exports.Monad = {
    ...exports.Pointed,
    ...exports.Chain,
};
/**
 * @category instances
 * @since 1.0.0
 */
exports.FromIO = {
    URI: exports.URI,
    fromIO: exports.fromIO,
};
/**
 * @category instances
 * @since 1.0.0
 */
exports.MonadIO = {
    ...exports.Monad,
    fromIO: exports.fromIO,
};
/**
 * Same as [`filter`](#filter), but passing also the index to the iterating function.
 *
 * @category filtering
 * @since 1.0.0
 */
const filterWithIndex = (predicateWithIndex) => (fa) => {
    let i = 0;
    return {
        *[Symbol.iterator]() {
            for (const a of fa) {
                if (predicateWithIndex(i++, a)) {
                    yield a;
                }
            }
        },
    };
};
exports.filterWithIndex = filterWithIndex;
/**
 * @category filtering
 * @since 1.0.0
 */
const filter = (predicate) => (fa) => ({
    *[Symbol.iterator]() {
        for (const a of fa) {
            if (predicate(a)) {
                yield a;
            }
        }
    },
});
exports.filter = filter;
/**
 * @category filtering
 * @since 1.0.0
 */
const filterMapWithIndex = (f) => (fa) => {
    let i = 0;
    return {
        *[Symbol.iterator]() {
            for (const a of fa) {
                const optionB = f(i++, a);
                if (O.isSome(optionB)) {
                    yield optionB.value;
                }
            }
        },
    };
};
exports.filterMapWithIndex = filterMapWithIndex;
/**
 * @category filtering
 * @since 1.0.0
 */
const filterMap = (f) => (0, exports.filterMapWithIndex)((_, a) => f(a));
exports.filterMap = filterMap;
/**
 * Compacts an Iterable of `Option`s discarding the `None` values and
 * keeping the `Some` values. It returns a new array containing the values of
 * the `Some` options.
 * @category filtering
 * @since 1.0.0
 */
exports.compact = 
/*#__PURE__*/ (0, exports.filterMap)(function_1.identity);
/**
 * @category filtering
 * @since 1.0.0
 */
const rights = (fa) => ({
    *[Symbol.iterator]() {
        for (const a of fa) {
            if (a._tag === "Right") {
                yield a.right;
            }
        }
    },
});
exports.rights = rights;
/**
 * @category filtering
 * @since 1.0.0
 */
const lefts = (ai) => ({
    *[Symbol.iterator]() {
        for (const a of ai) {
            if (a._tag === "Left") {
                yield a.left;
            }
        }
    },
});
exports.lefts = lefts;
/**
 * Creates a new `Iterable` removing duplicate elements, keeping the first occurrence of an element,
 * based on a `Eq<A>`.
 * @category filtering
 * @since 1.0.0
 */
const uniq = (E) => (fa) => {
    const uniques = [];
    return {
        *[Symbol.iterator]() {
            for (const a of fa) {
                if (uniques.every((o) => !E.equals(o, a))) {
                    uniques.push(a);
                    yield a;
                }
            }
        },
    };
};
exports.uniq = uniq;
/**
 * @category mapping
 * @since 1.0.0
 */
function transform(transform, flush) {
    return (fa) => ({
        *[Symbol.iterator]() {
            for (const a of fa) {
                const b = transform(a);
                if (O.isSome(b)) {
                    yield b.value;
                }
            }
            if (flush) {
                yield flush();
            }
        },
    });
}
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * @category combinators
 * @since 1.0.0
 */
exports.tap = (0, internal_1.dual)(2, (0, internal_1.tap)(exports.Chain));
/**
 * @category combinators
 * @since 1.0.0
 */
exports.tapIO = (0, internal_1.dual)(2, (0, internal_1.tapIO)(exports.FromIO, exports.Chain));
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @category do notation
 * @since 1.0.0
 */
exports.Do = (0, exports.of)({});
/**
 * @category do notation
 * @since 1.0.0
 */
exports.bind = chainable.bind(exports.Chain);
/**
 * @category do notation
 * @since 1.0.0
 */
exports.bindTo = (0, Functor_1.bindTo)(exports.Functor);
const _let = (0, Functor_1.let)(exports.Functor);
exports.let = _let;
/**
 * @category do notation
 * @since 1.0.0
 */
exports.apS = (0, Apply_1.apS)(exports.Apply);
/**
 * @category folding
 * @since 1.0.0
 */
const reduceWithIndex = (b, f) => (fa) => {
    let i = 0;
    let out = b;
    for (const a of fa) {
        out = f(i++, out, a);
    }
    return out;
};
exports.reduceWithIndex = reduceWithIndex;
/**
 * @category folding
 * @since 1.0.0
 */
const reduce = (b, f) => (0, exports.reduceWithIndex)(b, (_, b, a) => f(b, a));
exports.reduce = reduce;
/**
 * @category folding
 * @since 1.0.0
 */
const foldMapWithIndex = (M) => (f) => (fa) => (0, function_1.pipe)(fa, (0, exports.reduceWithIndex)(M.empty, (i, b, a) => M.concat(b, f(i, a))));
exports.foldMapWithIndex = foldMapWithIndex;
/**
 * @category folding
 * @since 1.0.0
 */
const foldMap = (M) => {
    const foldMapWithIndexM = (0, exports.foldMapWithIndex)(M);
    return (f) => foldMapWithIndexM((_, a) => f(a));
};
exports.foldMap = foldMap;
/**
 * @category conversions
 * @since 1.0.0
 */
const toArray = (fa) => Array.from(fa);
exports.toArray = toArray;
